// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AO20/16AO20_Linux_2.x.x.x_GSC_DN/driver/device.c $
// $Rev: 52768 $
// $Date: 2023-04-05 15:16:58 -0500 (Wed, 05 Apr 2023) $

// 16AO20: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model		Vendor	Device	SubVen	SubDev	type
	{ "16AO20",	0x10B5, 0x9080, 0x10B5, 0x3102,	GSC_DEV_TYPE_16AO20	},
	{ NULL }
};



//*****************************************************************************
static void _channels_compute(dev_data_t* dev)
{
	dev->cache.channels_max	= 20;

	switch ((dev->cache.gsc_for_32 >> 20) & 0x3)
	{
		default:
		case 0:		dev->cache.channel_mask	= 0x3F;
					dev->cache.channel_qty	= 6;
					break;

		case 1:		dev->cache.channel_mask	= 0xFFF;
					dev->cache.channel_qty	= 12;
					break;

		case 2:		dev->cache.channel_mask	= 0xFFFFF;
					dev->cache.channel_qty	= 20;
					break;
	}
}



//*****************************************************************************
static void _ext_pol_compute(dev_data_t* dev)
{
	if ((dev->cache.gsc_for_32 & 0xFFF) >= 0x005)
		dev->cache.ext_polarity	= 1;
	else
		dev->cache.ext_polarity	= 0;
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(AO20_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_acr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO20_GSC_ACR));
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO20_GSC_BCR));
		dev->vaddr.gsc_bor_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO20_GSC_BOR));
		dev->vaddr.gsc_csr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO20_GSC_CSR));
		dev->vaddr.gsc_for_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO20_GSC_FOR));
		dev->vaddr.gsc_odbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO20_GSC_ODBR));
		dev->vaddr.gsc_srr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AO20_GSC_SRR));

		// Data cache initialization.
		dev->cache.gsc_for_32		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_for_32);

		#define	NRATE_MIN			68
		#define	NRATE_MAX			0xFFFF

		dev->cache.auto_cal_ms		= 5000;
		dev->cache.fifo_size		= _256K;
		dev->cache.filter			= (dev->cache.gsc_for_32 >> 18) & 0x3;
		dev->cache.fref_default		= 30000000L;
		dev->cache.fsamp_max		= 30000000L / NRATE_MIN;	// 441,176 S/S
		dev->cache.fsamp_min		= 16000000L / NRATE_MAX;	//     244 S/S
		dev->cache.initialize_ms	= 3;
		dev->cache.nclk_mask		= 0x1FF;
		dev->cache.nclk_max			= 0x1FF;
		dev->cache.nclk_min			= 0;
		dev->cache.nrate_mask		= NRATE_MAX;
		dev->cache.nrate_max		= NRATE_MAX;
		dev->cache.nrate_min		= NRATE_MIN;
		dev->cache.range			= (dev->cache.gsc_for_32 >> 16) & 0x3;

		_channels_compute(dev);
		_ext_pol_compute(dev);

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_WRITE;
		ret	= gsc_dma_create(dev, dma, dma);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


